@@ -1,13 +1,16 @@
+import cgi
 import codecs
-from lxml import etree
-from lxml.cssselect import CSSSelector
+import operator
 import os
 import re
-import urllib
 import urlparse
+import urllib2
+
+import cssutils
+from lxml import etree
+from lxml.cssselect import CSSSelector
 
 
-__version__ = '1.10'
 __all__ = ['PremailerError', 'Premailer', 'transform']
 
 
@@ -18,7 +21,7 @@ class PremailerError(Exception):
 grouping_regex = re.compile('([:\-\w]*){([^}]+)}')
 
 
-def _merge_styles(old, new, class_=''):
+def merge_styles(old, new, class_=''):
     """
     if ::
       old = 'font-size:1px; color: red'
@@ -34,52 +37,61 @@ def _merge_styles(old, new, class_=''):
     Note: old could be something like '{...} ::first-letter{...}'
 
     """
-    news = {}
-    for k, v in [x.strip().split(':', 1) for x in new.split(';') if x.strip()]:
-        news[k.strip()] = v.strip()
+
+    def csstext_to_pairs(csstext):
+        parsed = cssutils.css.CSSVariablesDeclaration(csstext)
+        for key in parsed:
+            yield (key, parsed.getVariableValue(key))
+
+    new_keys = set()
+    news = []
+    for k, v in csstext_to_pairs(new):
+        news.append((k.strip(), v.strip()))
+        new_keys.add(k.strip())
 
     groups = {}
     grouped_split = grouping_regex.findall(old)
     if grouped_split:
         for old_class, old_content in grouped_split:
-            olds = {}
-            for k, v in [x.strip().split(':', 1) for
-                         x in old_content.split(';') if x.strip()]:
-                olds[k.strip()] = v.strip()
+            olds = []
+            for k, v in csstext_to_pairs(old_content):
+                olds.append((k.strip(), v.strip()))
             groups[old_class] = olds
     else:
-        olds = {}
-        for k, v in [x.strip().split(':', 1) for
-                     x in old.split(';') if x.strip()]:
-            olds[k.strip()] = v.strip()
+        olds = []
+        for k, v in csstext_to_pairs(old):
+            olds.append((k.strip(), v.strip()))
         groups[''] = olds
 
     # Perform the merge
-    merged = news
-    for k, v in groups.get(class_, {}).items():
-        if k not in merged:
-            merged[k] = v
+    relevant_olds = groups.get(class_, {})
+    merged = [style for style in relevant_olds if style[0] not in new_keys] + news
     groups[class_] = merged
 
     if len(groups) == 1:
-        return '; '.join(['%s:%s' % (k, v) for
-                          (k, v) in groups.values()[0].items()])
+        return '; '.join('%s:%s' % (k, v) for
+                          (k, v) in sorted(groups.values()[0]))
     else:
         all = []
         for class_, mergeable in sorted(groups.items(),
                                         lambda x, y: cmp(x[0].count(':'),
                                                          y[0].count(':'))):
             all.append('%s{%s}' % (class_,
-                                   '; '.join(['%s:%s' % (k, v)
+                                   '; '.join('%s:%s' % (k, v)
                                               for (k, v)
-                                              in mergeable.items()])))
-        return ' '.join([x for x in all if x != '{}'])
+                                              in mergeable)))
+        return ' '.join(x for x in all if x != '{}')
 
 
-_css_comments = re.compile(r'/\*.*?\*/', re.MULTILINE | re.DOTALL)
-_regex = re.compile('((.*?){(.*?)})', re.DOTALL | re.M)
-_semicolon_regex = re.compile(';(\s+)')
-_colon_regex = re.compile(':(\s+)')
+def make_important(bulk):
+    """makes every property in a string !important.
+    """
+    return ';'.join('%s !important' % p if not p.endswith('!important') else p
+                    for p in bulk.split(';'))
+
+
+_element_selector_regex = re.compile(r'(^|\s)\w')
+_cdata_regex = re.compile(r'\<\!\[CDATA\[(.*?)\]\]\>', re.DOTALL)
 _importants = re.compile('\s*!important')
 # These selectors don't apply to all elements. Rather, they specify
 # which elements to apply to.
@@ -90,12 +102,14 @@ class Premailer(object):
 
     def __init__(self, html, base_url=None,
                  preserve_internal_links=False,
-                 exclude_pseudoclasses=False,
+                 exclude_pseudoclasses=True,
                  keep_style_tags=False,
                  include_star_selectors=False,
                  remove_classes=True,
                  strip_important=True,
-                 external_styles=None):
+                 external_styles=None,
+                 method="html",
+                 base_path=None):
         self.html = html
         self.base_url = base_url
         self.preserve_internal_links = preserve_internal_links
@@ -109,20 +123,35 @@ class Premailer(object):
             external_styles = [external_styles]
         self.external_styles = external_styles
         self.strip_important = strip_important
+        self.method = method
+        self.base_path = base_path
 
-    def _parse_style_rules(self, css_body):
+    def _parse_style_rules(self, css_body, ruleset_index):
         leftover = []
         rules = []
-        css_body = _css_comments.sub('', css_body)
-        for each in _regex.findall(css_body.strip()):
-            __, selectors, bulk = each
-
-            bulk = _semicolon_regex.sub(';', bulk.strip())
-            bulk = _colon_regex.sub(':', bulk.strip())
-            if bulk.endswith(';'):
-                bulk = bulk[:-1]
-            for selector in [x.strip() for
-                             x in selectors.split(',') if x.strip()]:
+        rule_index = 0
+        # empty string
+        if not css_body:
+            return rules, leftover
+        sheet = cssutils.parseString(css_body)
+        for rule in sheet:
+            # ignore comment
+            if rule.type == rule.COMMENT:
+                continue
+            # handle media rule
+            if rule.type == rule.MEDIA_RULE:
+                leftover.append(rule)
+                continue
+            bulk = ';'.join(
+                u'{0}:{1}'.format(key, rule.style[key])
+                for key in rule.style.keys()
+            )
+            selectors = (
+                x.strip()
+                for x in rule.selectorText.split(',')
+                if x.strip() and not x.strip().startswith('@')
+            )
+            for selector in selectors:
                 if (':' in selector and self.exclude_pseudoclasses and
                     ':' + selector.split(':', 1)[1]
                         not in FILTER_PSEUDOSELECTORS):
@@ -132,11 +161,19 @@ class Premailer(object):
                 elif selector == '*' and not self.include_star_selectors:
                     continue
 
-                rules.append((selector, bulk))
+                # Crudely calculate specificity
+                id_count = selector.count('#')
+                class_count = selector.count('.')
+                element_count = len(_element_selector_regex.findall(selector))
+
+                specificity = (id_count, class_count, element_count, ruleset_index, rule_index)
+
+                rules.append((specificity, selector, bulk))
+                rule_index += 1
 
         return rules, leftover
 
-    def transform(self, pretty_print=True):
+    def transform(self, pretty_print=True, **kwargs):
         """change the self.html and return it with CSS turned into style
         attributes.
         """
@@ -144,8 +181,12 @@ class Premailer(object):
             return self.html
 
         parser = etree.HTMLParser()
-        tree = etree.fromstring(self.html.strip(), parser).getroottree()
+        stripped = self.html.strip()
+        tree = etree.fromstring(stripped, parser).getroottree()
         page = tree.getroot()
+        # lxml inserts a doctype if none exists, so only include it in
+        # the root if it was in the original html.
+        root = tree if stripped.startswith(tree.docinfo.doctype) else page
 
         if page is None:
             print repr(self.html)
@@ -157,40 +198,72 @@ class Premailer(object):
         ##
 
         rules = []
+        index = 0
 
-        for style in CSSSelector('style')(page):
-            css_body = etree.tostring(style)
-            css_body = css_body.split('>')[1].split('</')[0]
-            these_rules, these_leftover = self._parse_style_rules(css_body)
+        for element in CSSSelector('style,link[rel~=stylesheet]')(page):
+            # If we have a media attribute whose value is anything other than
+            # 'screen', ignore the ruleset.
+            media = element.attrib.get('media')
+            if media and media != 'screen':
+                continue
+
+            is_style = element.tag == 'style'
+            if is_style:
+                css_body = element.text
+            else:
+                href = element.attrib.get('href')
+                if not href:
+                    continue
+                css_body = self._load_external(href)
+
+            these_rules, these_leftover = self._parse_style_rules(css_body, index)
+            index += 1
             rules.extend(these_rules)
 
-            parent_of_style = style.getparent()
+            parent_of_element = element.getparent()
             if these_leftover:
-                style.text = '\n'.join(['%s {%s}' % (k, v) for
-                                        (k, v) in these_leftover])
-            elif not self.keep_style_tags:
-                parent_of_style.remove(style)
+                if is_style:
+                    style = element
+                else:
+                    style = etree.Element('style')
+                    style.attrib['type'] = 'text/css'
+
+                lines = []
+                for item in these_leftover:
+                    if isinstance(item, tuple):
+                        k, v = item
+                        lines.append('%s {%s}' % (k, make_important(v)))
+                    # media rule
+                    else:
+                        for rule in item.cssRules:
+                            for key in rule.style.keys():
+                                rule.style[key] = (rule.style.getPropertyValue(key, False), '!important')
+                        lines.append(item.cssText)
+                style.text = '\n'.join(lines)
+                if self.method == 'xml':
+                    style.text = etree.CDATA(style.text)
+
+                if not is_style:
+                    element.addprevious(style)
+                    parent_of_element.remove(element)
+
+            elif not self.keep_style_tags or not is_style:
+                parent_of_element.remove(element)
 
         if self.external_styles:
             for stylefile in self.external_styles:
-                print stylefile
-                if stylefile.startswith('http://'):
-                    css_body = urllib.urlopen(stylefile).read()
-                elif os.path.exists(stylefile):
-                    try:
-                        f = codecs.open(stylefile)
-                        css_body = f.read()
-                    finally:
-                        f.close()
-                else:
-                    raise ValueError(u"Could not find external style: %s" %
-                                     stylefile)
-                these_rules, these_leftover = self._parse_style_rules(css_body)
+                css_body = self._load_external(stylefile)
+                these_rules, these_leftover = self._parse_style_rules(css_body, index)
+                index += 1
                 rules.extend(these_rules)
 
+        # rules is a tuple of (specificity, selector, styles), where specificity is a tuple
+        # ordered such that more specific rules sort larger.
+        rules.sort(key=operator.itemgetter(0))
+
         first_time = []
         first_time_styles = []
-        for selector, style in rules:
+        for __, selector, style in rules:
             new_selector = selector
             class_ = ''
             if ':' in selector:
@@ -206,14 +279,11 @@ class Premailer(object):
             for item in sel(page):
                 old_style = item.attrib.get('style', '')
                 if not item in first_time:
-                    if old_style:
-                        new_style = _merge_styles(style, old_style, class_)
-                    else:
-                        new_style = _merge_styles(old_style, style, class_)
+                    new_style = merge_styles(old_style, style, class_)
                     first_time.append(item)
                     first_time_styles.append((item, old_style))
                 else:
-                    new_style = _merge_styles(old_style, style, class_)
+                    new_style = merge_styles(old_style, style, class_)
                 item.attrib['style'] = new_style
                 self._style_to_basic_html_attributes(item, new_style,
                                                      force=True)
@@ -223,7 +293,7 @@ class Premailer(object):
             old_style = item.attrib.get('style', '')
             if not inline_style:
                 continue
-            new_style = _merge_styles(old_style, inline_style, class_)
+            new_style = merge_styles(old_style, inline_style, class_)
             item.attrib['style'] = new_style
             self._style_to_basic_html_attributes(item, new_style, force=True)
 
@@ -243,15 +313,55 @@ class Premailer(object):
                     if attr == 'href' and self.preserve_internal_links \
                            and parent.attrib[attr].startswith('#'):
                         continue
+                    if not self.base_url.endswith('/'):
+                        self.base_url += '/'
                     parent.attrib[attr] = urlparse.urljoin(self.base_url,
-                                                           parent.attrib[attr])
+                        parent.attrib[attr].lstrip('/'))
 
-        out = etree.tostring(page, pretty_print=pretty_print).replace(
-            '<head/>', '<head></head>')
+        kwargs.setdefault('method', self.method)
+        kwargs.setdefault('pretty_print', pretty_print)
+        out = etree.tostring(root, **kwargs)
+        if self.method == 'xml':
+            out = _cdata_regex.sub(lambda m: '/*<![CDATA[*/%s/*]]>*/' % m.group(1), out)
         if self.strip_important:
             out = _importants.sub('', out)
         return out
 
+    def _load_external_url(self, url):
+        r = urllib2.urlopen(url)
+        _, params = cgi.parse_header(r.headers.get('Content-Type', ''))
+        encoding = params.get('charset', 'utf-8')
+        return r.read().decode(encoding)
+
+    def _load_external(self, url):
+        """loads an external stylesheet from a remote url or local path
+        """
+        if url.startswith('//'):
+            # then we have to rely on the base_url
+            if self.base_url and 'https://' in self.base_url:
+                url = 'https:' + url
+            else:
+                url = 'http:' + url
+
+        if url.startswith('http://') or url.startswith('https://'):
+            css_body = self._load_external_url(url)
+        else:
+            stylefile = url
+            if not os.path.isabs(stylefile):
+                stylefile = os.path.abspath(
+                    os.path.join(self.base_path or '', stylefile)
+                )
+            if os.path.exists(stylefile):
+                with codecs.open(stylefile, encoding='utf-8') as f:
+                    css_body = f.read()
+            elif self.base_url:
+                url = urlparse.urljoin(self.base_url, url)
+                return self._load_external(url)
+            else:
+                raise ValueError(u"Could not find external style: %s" %
+                                 stylefile)
+        return css_body
+
     def _style_to_basic_html_attributes(self, element, style_content,
                                         force=False):
         """given an element and styles like
